"""Integration with [Pandas](https://pandas.pydata.org/) to aid in data cleaning dataframes."""

import pandas
import math
from sympy import Symbol, lambdify
from pandas.api.types import (
    is_unsigned_integer_dtype,
    is_integer_dtype,
    is_float_dtype,
    is_complex_dtype,
)
from functools import cache, reduce
from itertools import combinations

from symconstraints import Constraints, Validation, Imputation
from sympy.logic.boolalg import Boolean
import numpy as np
from operator import or_


def symbols(
    df: pandas.DataFrame, symbol_list: str | list[str], **kwargs
) -> Symbol | list[Symbol]:
    """Return SymPy symbols with assumptions inferred from the dataframe dtypes.

    Currently, it infers the following:

    * Unsigned integer dtypes are inferred to be nonnegative integers
    * Float dtypes are inferred to be real numbers
    * Complex dtypes are inferred to be complex numbers

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe to infer symbol assumptions from.
    symbol_list : str | list[str]
        List of columns, can be represented as a space separated string.
    **kwargs: dict, optional
        Extra arguments to be passed to `sympy.Symbol`

    Returns
    -------
    Symbol | list[Symbol]
        SymPy symbols with assumptions inferred from the dataframe dtypes, corresponding to each column given.
        Returns a list if multiple columns are given, or a single symbol if a signle column is given.

    Raises
    ------
    ValueError
        Raises a ValueError if a column is not found or an unsupported dtype is found.

    Examples
    --------
    >>> import pandas as pd
    >>> from symconstraints.pandas import symbols
    >>> df = pd.DataFrame({
    ...    'Level': [1],
    ...    'Width': [5.3],
    ...    'Height': [7.6],
    ...    'Voltage': [5+3j]
    ... }).astype({'Level': 'uint8'})
    >>> level, width, voltage = symbols(df, 'Level Width Voltage')
    >>> level.is_nonnegative, level.is_integer
    (True, True)
    >>> width.is_integer, width.is_real
    (None, True)
    >>> voltage.is_real, voltage.is_complex
    (None, True)
    """
    result = []
    symbols = symbol_list if isinstance(symbol_list, list) else symbol_list.split()
    for symbol in symbols:
        symbol_dtype = df.dtypes.get(symbol)
        if symbol_dtype is None:
            raise ValueError(
                f"Column {symbol} does not exist in the dataframe provided. Available columns are {list(df.columns)}."
            )

        assumptions = (
            {"integer": True, "nonnegative": True}
            if is_unsigned_integer_dtype(symbol_dtype)
            else {"integer": True}
            if is_integer_dtype(symbol_dtype)
            else {"real": True}
            if is_float_dtype(symbol_dtype)
            else {"complex": True}
            if is_complex_dtype(symbol_dtype)
            else None
        )

        if assumptions is None:
            raise ValueError(
                f"Unsupported data type {symbol_dtype} in column {symbol}."
            )

        result.append(Symbol(symbol, **assumptions, **kwargs))

    match len(result):
        case 0:
            raise ValueError("No symbols given.")
        case 1:
            return result[0]

    return result


@cache
def _lambdify(*args, **kwargs):
    return lambdify(*args, **kwargs, modules=["numpy"])


def check(
    constraints: Constraints | Validation, df: pandas.DataFrame
) -> pandas.DataFrame:
    """Return a table checking all the validations provided.

    Parameters
    ----------
    constraints : Constraints | Validation
        `Constraints` or `Validation` to use for checking.
    df : pandas.DataFrame
        Dataframe to check

    Returns
    -------
    pandas.DataFrame
        Returns a dataframe showing the result of all the validations for each row in the dataframe.
        Each column in the dataframe corresponds to each validation operation provided. Each row
        corresponds to each row in the original dataframe. A result of 1 is shown for successful validation,
        0 for an unsuccessful validation, and NaN for validations that can't be computed due to missing values.
        The result dtype is `float16`.

        If a `Constraints` object is provided, the columns are `pandas.MultiIndex`, where the top level is a set of
        columns with all validations concerning those sets of columns under it.

    Examples
    --------
    >>> import pandas as pd
    >>> from symconstraints.pandas import symbols, check
    >>> from symconstraints import Constraints
    >>> from sympy import Eq
    >>> df = pd.DataFrame({
    ...     'A': [5,6,8,9],
    ...     'B': [3,5,90,None],
    ...     'C': [14, 30, None, None]
    ... }, dtype=float)
    >>> A, B, C = symbols(df, ['A', 'B', 'C'])
    >>> constraints = Constraints([A > B, Eq(C, B*A)])
    >>> check(constraints, df)
    # Order may differ
      (A, B)  (C, A, B)  (C, B)  (C, A)
       A > B Eq(C, A*B) B < C/B A > C/A
    0    1.0        0.0     1.0     1.0
    1    1.0        1.0     1.0     1.0
    2    0.0        NaN     NaN     NaN
    3    NaN        NaN     NaN     NaN
    """
    if isinstance(constraints, Validation):
        columns = tuple(constraints.keys)
        index: list[Boolean] = list(constraints.operations)
        columns_str = [str(column) for column in columns]

        relevant_indices = ~df[columns_str].isna().any(axis="columns")
        if not isinstance(relevant_indices, pandas.Series):
            raise RuntimeError(
                f"unexpected type for relevant_indices: {type(relevant_indices)}"
            )

        relevant_df = df[relevant_indices]
        # Return a dataframe if the output is a series
        return pandas.DataFrame(
            pandas.concat(
                [
                    pandas.concat(
                        dict(
                            (
                                operation,
                                _lambdify(columns, operation)(
                                    *(relevant_df[str(column)] for column in columns)
                                ).astype("float16"),
                            )
                            for operation in index
                        ),
                        axis="columns",
                    ),
                    pandas.concat(
                        dict(
                            (
                                operation,
                                pandas.Series(
                                    math.nan,
                                    index=relevant_indices.index[~relevant_indices],
                                    copy=True,
                                ),
                            )
                            for operation in index
                        ),
                        axis="columns",
                    ),
                ],
                axis="rows",
            )
        )

    if isinstance(constraints, Constraints):
        return pandas.concat(
            dict(
                (validation.keys, check(validation, df))
                for validation in constraints.validations
            ),
            axis="columns",
        )

    raise ValueError(f"Invalid constraints given: {constraints}")


def set_invalid_all(
    check_result: pandas.DataFrame, df: pandas.DataFrame, fill=math.nan
) -> pandas.DataFrame:
    """Replace all possible invalid values in the dataframe to a set value.

    This replaces values in the dataframe that could possibly be invalid under the given constraints. This might help
    get rid of outlier data within the dataframe.

    The input dataframe is copied and is not edited in-place.

    Parameters
    ----------
    check_result : pandas.DataFrame
        Check result returned by `check_result`
    df : pandas.DataFrame
        Dataframe to edit
    fill : Any, optional
        The set value to replace invalid values.

    Returns
    -------
    pandas.DataFrame
        Dataframe with replaced values.

    Examples
    --------
    >>> import pandas as pd
    >>> from symconstraints import Constraints
    >>> from symconstraints.pandas import symbols, check, set_invalid_all
    >>> from sympy import Eq
    >>> df = pd.DataFrame(
    ...    {
    ...         "height": [5, 6, 8, 9],
    ...         "width": [3, 5, 90, None],
    ...         "area": [14, 30, None, 18],
    ...     },
    ...     dtype=float,
    ... )
    >>> height, width, area = symbols(df, ["height", "width", "area"])
    >>> constraints = Constraints([height > width, Eq(area, width * height)])
    >>> check_result = check(constraints, df)
    >>> set_invalid_all(check_result, df)
       height  width  area
    0     NaN    NaN   NaN
    1     6.0    5.0  30.0
    2     NaN    NaN   NaN
    3     9.0    NaN  18.0
    """
    result = df.copy()
    if check_result.empty:
        return result
    if check_result.columns.nlevels == 1:
        column_item = check_result.columns.item()
        if not isinstance(column_item, Boolean):
            raise ValueError(
                f"Invalid check result given. It has a column {column_item} of type: {type(column_item)}"
            )
        result.loc[
            ~check_result.fillna(1.0).all(axis="columns"),
            [str(symbol) for symbol in column_item.free_symbols],
        ] = fill

        return result

    keysets = check_result.columns.get_level_values(0)

    invalid_keysets = [keys for keys in keysets if not isinstance(keys, frozenset)]
    if len(invalid_keysets) > 0:
        raise ValueError(f"Found invalid columns in check result: {invalid_keysets}")

    for keys in keysets:
        result.loc[
            ~check_result[keys].fillna(1.0).all(axis="columns"),
            [str(key) for key in keys],
        ] = fill

    return result


@cache
def _set_cover(subsets_buffer: bytes, shape: tuple[int, int]) -> list[int]:
    if shape[0] == 0 or shape[1] == 0:
        return []

    subsets = np.frombuffer(subsets_buffer, dtype=bool).reshape(shape)
    for symbol_amount in range(1, subsets.shape[1]):
        for selection in combinations(range(subsets.shape[1]), symbol_amount):
            if subsets[:, selection].any(axis=1).all():
                return list(selection)
    return list(range(subsets.shape[1]))


def set_invalid_min(
    check_result: pandas.DataFrame,
    df: pandas.DataFrame,
    fill=math.nan,
    priority: list[Symbol] | None = None,
) -> pandas.DataFrame:
    """Replace the minimum amount of possible invalid values in the dataframe to a set value.

    Similar to `set_invalid_all`, this replaces values in the dataframe that could possibly be invalid under the given constraints.
    However it does not replace all the values, instead it tries to replace the minimum amount of values in each row such that
    it satisfies all the constraints. This might help get rid of outlier data within the dataframe, while also keeping more of the values
    in the dataset unchanged.

    The input dataframe is copied and is not edited in-place.

    Parameters
    ----------
    check_result : pandas.DataFrame
        Check result returned by `check_result`
    df : pandas.DataFrame
        Dataframe to edit
    fill : Any, optional
        The set value to replace invalid values.
    priority : list[Symbol] | None, optional
        In case of a tie in the decision to set which column, refer to this priority to decide which
        column to to set first. Priority is ordered from most preferred to be set to least preferred.

    Returns
    -------
    pandas.DataFrame
        Dataframe with replaced values.

    Examples
    --------
    >>> import pandas as pd
    >>> from symconstraints import Constraints
    >>> from symconstraints.pandas import symbols, check, set_invalid_min
    >>> from sympy import Eq
    >>> df = pd.DataFrame(
    ...    {
    ...        "height": [5, 6, 8, 9, 7],
    ...        "width": [3, 5, 90, None, 8],
    ...        "depth": [None, 2, 10, 5, 5],
    ...        "area": [14, 30, 10, None, 35],
    ...        "volume": [None, 60, 100, 30, None],
    ...    },
    ...    dtype=float,
    ... )
    >>> height, width, area, depth, volume = symbols(df, ["height", "width", "area", "depth", "volume"])
    >>> constraints = Constraints([
    ...     Eq(area, height * width),
    ...     Eq(volume, area * depth),
    ...     height > width,
    ...     width > depth,
    ... ])
    >>> check_result = check(constraints, df)
    >>> set_invalid_min(check_result, df, priority=[volume, area, depth, width, height])
       height  width  depth  area  volume
    0     5.0    3.0    NaN   NaN     NaN
    1     6.0    5.0    2.0  30.0    60.0
    2     NaN   90.0   10.0   NaN   100.0
    3     9.0    NaN    5.0   NaN    30.0
    4     7.0    NaN    5.0   NaN     NaN
    """
    validation_sets = check_result.columns.get_level_values(0).unique()
    symbols = reduce(or_, validation_sets, frozenset())
    symbols_str = [str(s) for s in symbols]

    if priority is not None:
        priority_indices = dict(
            (str(value), index) for index, value in enumerate(priority)
        )
        symbols_str.sort(key=lambda symbol: priority_indices.get(symbol, math.inf))

    # Finding the minimum number of symbols to replace is equivalent to the set cover problem.
    check_sets = check_result.groupby(level=0, axis="columns").all()
    check_sets.columns = check_sets.columns.get_level_values(0)

    set_to_symbols = pandas.DataFrame(
        [
            dict(
                (symbol, symbol in frozenset(str(s) for s in subset))
                for symbol in symbols_str
            )
            for subset in validation_sets
        ],
        index=validation_sets,
    )

    set_cover_results = pandas.DataFrame(columns=[*validation_sets, *symbols_str])

    for _, selection in check_sets.drop_duplicates().iterrows():
        invalid_sets = selection.index[selection == 0.0]
        if not isinstance(invalid_sets, pandas.Index):
            raise ValueError(
                f"Unexpected value found when fetching invalid sets index: {invalid_sets}"
            )
        invalid_set_to_symbols = set_to_symbols.loc[invalid_sets, :]
        invalid_set_to_symbols = invalid_set_to_symbols[
            invalid_set_to_symbols.any(axis="columns")
        ]
        invalid_symbols = list(invalid_set_to_symbols.columns)
        invalid_set_to_symbols_numpy = invalid_set_to_symbols.to_numpy()
        symbols_to_remove = set(
            invalid_symbols[index]
            for index in _set_cover(
                invalid_set_to_symbols_numpy.tobytes(),
                invalid_set_to_symbols_numpy.shape,
            )
        )
        set_cover_results.loc[len(set_cover_results)] = pandas.concat(
            [
                selection,
                pandas.Series(
                    dict(
                        (symbol, symbol in symbols_to_remove) for symbol in symbols_str
                    )
                ),
            ]
        )

    to_set = check_sets.merge(set_cover_results, how="left")[symbols_str]
    to_set.index = check_sets.index  # I don't know why I have to do this.

    return df.where(~to_set, fill)


def impute(
    constraints: Constraints | Imputation, df: pandas.DataFrame
) -> pandas.DataFrame:
    """Impute the dataframe under the given constraints.

    This returns a copy of the dataframe with all NA values replaced with values inferred from
    imputation operations given in the constraints. This assumes that all the values are valid,
    so it is recommended that the dataframe be checked and that all of its invalid values are
    removed via `check` and `set_invalid_all`.

    The input dataframe is not edited in-place.

    Parameters
    ----------
    constraints : Constraints | Imputation
        `Constraints` or `Imputation` to use for imputation.
    df : pandas.DataFrame
        Dataframe to impute.

    Returns
    -------
    pandas.DataFrame
        Imputed dataframe.

    Examples
    --------
    >>> import pandas as pd
    >>> from symconstraints import Constraints
    >>> from symconstraints.pandas import symbols, check, set_invalid_all, impute
    >>> from sympy import Eq
    >>> df = pd.DataFrame(
    ...    {
    ...         "height": [5, 6, 8, 9],
    ...         "width": [3, 5, 7, None],
    ...         "area": [14, 30, None, 18],
    ...     },
    ...     dtype=float,
    ... )
    >>> height, width, area = symbols(df, ["height", "width", "area"])
    >>> constraints = Constraints([height > width, Eq(area, width * height)])
    >>> check_result = check(constraints, df)
    >>> df = set_invalid_all(check_result, df)
    >>> df
       height  width  area
    0     NaN    NaN   NaN
    1     6.0    5.0  30.0
    2     8.0    7.0   NaN
    3     9.0    NaN  18.0
    >>> imputed_df = impute(constraints, df)
    >>> imputed_df
       height  width  area
    0     NaN    NaN   NaN
    1     6.0    5.0  30.0
    2     8.0    7.0  56.0
    3     9.0    2.0  18.0
    """
    if isinstance(constraints, Imputation):
        result = df.copy()

        columns = tuple(constraints.keys)
        columns_str = [str(column) for column in columns]

        result.loc[
            result[columns_str].notna().all(axis="columns")
            & result[str(constraints.target_key)].isna(),
            str(constraints.target_key),
        ] = _lambdify(columns, constraints.operation)(
            *(result[column_str] for column_str in columns_str)
        )

        return result

    if isinstance(constraints, Constraints):
        result = df.copy()

        for imputation in constraints.imputations:
            columns = tuple(imputation.keys)
            columns_str = [str(column) for column in columns]

            result.loc[
                result[columns_str].notna().all(axis="columns")
                & result[str(imputation.target_key)].isna(),
                str(imputation.target_key),
            ] = _lambdify(columns, imputation.operation)(
                *(result[column_str] for column_str in columns_str)
            )

        return result

    raise ValueError(f"Invalid constraints given: {constraints}")



"""Basic implementations of operations to the standard python mapping (dict, defaultdict, etc.)."""

from sympy.logic.boolalg import Boolean
from symconstraints import Validation, Constraints, Imputation

from collections.abc import Mapping
from typing import Any

StringMap = Mapping[str, Any]
"""Any generic map that has a string key, such as dict, defaultdict, Counter, etc."""


class ValidationError(Exception):
    """Validation error.

    Attributes
    ----------
    values
        Mapping values relevant to the validation operations.

    unsatisfied_booleans : list[Boolean]
        List of equalities/inequalities where these values do not satisfy.
    """

    values: StringMap
    unsatisfied_booleans: list[Boolean]

    def __init__(self, values: StringMap, unsatisfied_booleans: list[Boolean]):
        super().__init__(
            f"Mapping {values} is invalid due to not satisfying [{', '.join(str(op) for op in unsatisfied_booleans)}]"
        )

        self.values = values
        self.unsatisfied_booleans = unsatisfied_booleans


class ConstraintsValidationError(Exception):
    """Constraints validation error.

    Attributes
    ----------
    validation_errors : list[ValidationError]
        List of errors for each validations in the constraints.
    """

    validation_errors: list[ValidationError]

    def __init__(self, validation_errors: list[ValidationError]):
        super().__init__(
            "Mapping is invalid due to:"
            + "".join(
                f"\n- {validation_error}" for validation_error in validation_errors
            )
        )

        self.validation_errors = validation_errors


def validate(constraints: Constraints | Validation, mapping: StringMap):
    """Validate mapping via a validation or constraints.

    Parameters
    ----------
    constraints : Constraints | Validation
        Constraints or validation to use for validation.
    mapping : AnyValueMap
        Input to validate

    Raises
    ------
    ValidationError
        Raised when the mapping data given is invalid under constraints of type `Validation`
    ConstraintsValidationError
        Raised when the mapping data given is invalid under constraints of type `Constraints`

    Examples
    --------
    >>> from symconstraints import Constraints, symbols
    >>> from symconstraints.mapping import validate
    >>> a, b, c = symbols('a b c')
    >>> constraints = Constraints([a < b, b < c])
    >>> validate(constraints, {'a': 1, 'b': 2})
    >>> # Nothing happens, data is valid
    >>> try:
    ...   validate(constraints, {'a': 4, 'c': 1})
    ... except ConstraintsValidationError as e:
    ...   print(e)
    Mapping is invalid due to:
    - Mapping {'a': 4, 'c': 1} is invalid due to not satisfying [a < c]
    """
    if isinstance(constraints, Validation):
        values = dict((str(k), mapping.get(str(k))) for k in constraints.keys)

        if any(mapping.get(str(key)) is None for key in constraints.keys):
            return

        values_subs = [(k, mapping[str(k)]) for k in constraints.keys]

        unsatisfied_expressions: list[Boolean] = [
            operation
            for operation in constraints.operations
            if not operation.subs(values_subs)
        ]

        if len(unsatisfied_expressions) > 0:
            raise ValidationError(
                values,
                unsatisfied_expressions,
            )
    elif isinstance(constraints, Constraints):
        errors: list[ValidationError] = []
        for validation in constraints.validations:
            try:
                validate(validation, mapping)
            except ValidationError as e:
                errors.append(e)

        if len(errors) > 0:
            raise ConstraintsValidationError(errors)
    else:
        raise ValueError(f"Invalid constraints given: {constraints}")


def impute(constraints: Imputation | Constraints, mapping: StringMap) -> dict[str, Any]:
    """Impute mapping via a validation or constraints object.

    Parameters
    ----------
    constraints : Imputation | Constraints
        Constraints or Imputation object to use for imputation.
    mapping : StringMap
        Input to impute.

    Returns
    -------
    dict[str, Any]
        The imputed mapping as a dictionary.

    Examples
    --------
    >>> from sympy import Eq
    >>> from symconstraints import symbols, Constraints
    >>> from symconstraints.mapping import impute
    >>> a, b, c, d = symbols("a b c d")
    >>> constraints = Constraints([Eq(a, 2 * b + c), c < b, Eq(d, a * c)])
    >>> impute(constraints, {"b": 10, "c": 3})
    {'a': 23, 'b': 10, 'c': 3, 'd': 69}
    """
    if isinstance(constraints, Imputation):
        if mapping.get(str(constraints.target_key)) is None and all(
            mapping.get(str(key)) is not None for key in constraints.keys
        ):
            return {
                **mapping,
                str(constraints.target_key): constraints.operation.subs(
                    [(key, mapping[str(key)]) for key in constraints.keys]
                ),
            }
        return dict(mapping)
    elif isinstance(constraints, Constraints):
        result = {**mapping}
        for imputation in constraints.imputations:
            result = impute(imputation, result)
        return result
    else:
        raise ValueError(f"Invalid constraints given: {constraints}")



"""Tools to create constraints."""

from __future__ import annotations

from collections import defaultdict
from itertools import combinations
from typing import TYPE_CHECKING
from warnings import warn
from dataclasses import dataclass
from functools import reduce

import sympy
from sympy import (
    And,
    Dummy,
    Eq,
    Basic,
    Ge,
    Gt,
    Interval,
    Le,
    Lt,
    Or,
    S,
    Symbol,
    oo,
    simplify_logic,
    solveset,
    FiniteSet,
    Expr,
    Intersection,
    Union,
    ConditionSet,
    roots,
    reduce_inequalities,
    Q,
    ask,
    Pow,
    Poly,
)
from sympy.polys.polyerrors import UnsolvableFactorError
from sympy.logic.boolalg import Boolean

if TYPE_CHECKING:
    from typing import Iterable, Literal

from symconstraints.operation import Imputation, Validation

_assumption_domain = {
    "real": S.Reals,
    "complex": S.Complexes,
    "integer": S.Integers,
    "negative": Interval(-oo, 0),
    "positive": Interval(0, oo),
}


def _get_symbol_domain(symbol):
    result = S.Complexes
    for assumption, is_assumption_true in symbol.assumptions0.items():
        domain = _assumption_domain.get(assumption)
        if domain is not None:
            result = (
                result.intersect(domain)
                if is_assumption_true
                else domain.complement(result)
            )
    return result


class _DummyRelation:
    rel: Literal["<", ">", "=", "!"]
    expr: Basic
    strict: bool

    def __init__(self, relation: Basic, dummy: Dummy):
        relation = reduce_inequalities(relation, dummy)
        self.strict = isinstance(relation, (Lt, Gt))
        if isinstance(relation, Eq):
            self.rel = "="
            self.expr = relation.rhs if relation.lhs == dummy else relation.lhs
        elif isinstance(relation, Ge | Gt | Le | Lt):
            if relation.gts == dummy:
                self.rel = ">"
                self.expr = relation.lts
            else:
                self.rel = "<"
                self.expr = relation.gts

        if self.rel is None:
            raise ValueError(f"Could not analyze relation {relation}")


def _is_even_root(expr: _DummyRelation):
    return isinstance(expr.expr, Pow) and ask(Q.even(1 / expr.expr.args[1]))


def _and_dummy_to_constraints(and_relation: And, dummy: Dummy) -> set[Boolean]:
    constraints: set[Boolean] = set()
    useful_relations = []
    for rel in and_relation.args:
        try:
            useful_relations.append(_DummyRelation(rel, dummy))
        except ValueError as e:
            warn(str(e))

    for relation1, relation2 in combinations(useful_relations, 2):
        strict = relation1.strict or relation2.strict
        match (relation1.rel, relation2.rel):
            case ("=", "="):
                constraints.add(Eq(relation1.expr, relation2.expr))
            case (">" | "=", "<" | "="):
                if ask(Eq(relation1.expr, -relation2.expr)) and (
                    _is_even_root(relation1) or _is_even_root(relation2)
                ):
                    # This is a workaround to a bug where SymPy doesn't recognize the square roots are always nonnegative
                    # in the real domain
                    continue

                constraints.add((Lt if strict else Le)(relation1.expr, relation2.expr))
            case ("<" | "=", ">" | "="):
                if ask(Eq(relation1.expr, -relation2.expr)) and (
                    _is_even_root(relation1) or _is_even_root(relation2)
                ):
                    # This is a workaround to a bug where SymPy doesn't recognize the square roots are always nonnegative
                    # in the real domain
                    continue

                constraints.add((Gt if strict else Ge)(relation1.expr, relation2.expr))

    return constraints


def _get_basic_symbols(basic: Basic):
    return frozenset(
        symbol for symbol in basic.free_symbols if isinstance(symbol, Symbol)
    )


@dataclass(frozen=True)
class _InferredSet:
    inferred_set: sympy.Set
    inferred_by: Boolean


@dataclass(frozen=True)
class _InferredConstraint:
    constraint: Boolean
    inferred_by: frozenset[Boolean]


def symbols(symbols_str: str | Iterable[str], *, real=True, **kwargs):
    """Define SymPy symbols.

    This is equivalent to the sympy.symbols function, but symbols are defined in the real domain by default.

    Parameters
    ----------
    symbols_str : str | Iterable[str]
        Space seperated set of symbols or an array of symbols
    real : bool, optional
        Assume symbols are real numbers, by default True
    **kwargs
        Other keyword arguments are passed to the sympy.symbols functions
    """
    return sympy.symbols(symbols_str, real=real, **kwargs)


class Constraints:
    """Creates a set of validation and imputation operations from mathematical SymPy expressions.

    Examples
    --------
    Get all possible validations and imputations, including inferred ones

    >>> from sympy import Eq
    >>> from symconstraints import Constraints, symbols
    >>> a, b, c = symbols('a b c')
    >>> # a=b+c, c<b+3
    >>> constraints = Constraints([Eq(a, 2 * b), c < b + 3])
    >>> for validation in constraints.validations:
    ...     print(validation)
    ...
    Validation: (b, a) => [Eq(a, 2*b)] inferred by (Eq(a, 2*b))
    Validation: (c, b) => [b - c > -3] inferred by (c < b + 3)
    Validation: (c, a) => [a/2 > c - 3] inferred by (c < b + 3, Eq(a, 2*b))
    >>> for imputation in constraints.imputations:
    ...     print(imputation)
    ...
    Imputation: (b) => a = 2*b inferred by (Eq(a, 2*b))
    Imputation: (a) => b = a/2 inferred by (Eq(a, 2*b))
    """

    _validations: list[Validation]
    _imputations: list[Imputation]

    def __init__(self, constraints: Iterable[Boolean]):
        """Create constraints.

        Parameters
        ----------
        constraints : Iterable[Boolean]
            List of SymPy Boolean expressions
        """
        self._symbol_to_sets: defaultdict[Symbol, set[_InferredSet]] = defaultdict(set)
        self._symbols_to_constraints: defaultdict[
            frozenset[Symbol], set[_InferredConstraint]
        ] = defaultdict(set)
        self._imputations = []

        for constraint in constraints:
            simplified_constraint = simplify_logic(constraint, form="cnf", force=True)
            if isinstance(simplified_constraint, And):
                for arg in simplified_constraint.args:
                    if isinstance(arg, Boolean):
                        self._add_constraint(arg, constraint)
            else:
                self._add_constraint(simplified_constraint, constraint)

        for symbol, symbol_sets in self._symbol_to_sets.items():
            for inferred_set1, inferred_set2 in combinations(symbol_sets, 2):
                set1, inferred_by1 = (
                    inferred_set1.inferred_set,
                    inferred_set1.inferred_by,
                )
                set2, inferred_by2 = (
                    inferred_set2.inferred_set,
                    inferred_set2.inferred_by,
                )
                inferred_by = frozenset([inferred_by1, inferred_by2])
                dummy = Dummy(**symbol.assumptions0)
                dummy_relation = simplify_logic(
                    set1.intersect(set2).as_relational(dummy), form="dnf"
                )
                if isinstance(dummy_relation, Or):
                    and_operations: list[Boolean] = []
                    for arg in dummy_relation.args:
                        if isinstance(arg, And):
                            and_operations.append(
                                And(*_and_dummy_to_constraints(arg, dummy))
                            )
                        else:
                            and_operations = []
                            break
                    if len(and_operations) > 0:
                        simplified_constraint = Or(*and_operations)
                        self._symbols_to_constraints[
                            _get_basic_symbols(simplified_constraint)
                        ].add(_InferredConstraint(simplified_constraint, inferred_by))
                elif isinstance(dummy_relation, And):
                    for simplified_constraint in _and_dummy_to_constraints(
                        dummy_relation, dummy
                    ):
                        constraint_symbols = _get_basic_symbols(simplified_constraint)
                        self._symbols_to_constraints[constraint_symbols].add(
                            _InferredConstraint(simplified_constraint, inferred_by)
                        )
                        for constraint_symbol in constraint_symbols:
                            constraint_symbol_set = solveset(
                                simplified_constraint,
                                constraint_symbol,
                                domain=_get_symbol_domain(constraint_symbol),
                            )
                            self._add_possible_imputation_from_set(
                                constraint_symbol_set, constraint_symbol, inferred_by
                            )

        self._validations = [
            Validation(
                frozenset(symbols),
                frozenset(
                    inferred_constraint.constraint
                    for inferred_constraint in inferred_constraints
                ),
                inferred_by=reduce(
                    lambda a, b: a | b.inferred_by, inferred_constraints, frozenset()
                ),
            )
            for symbols, inferred_constraints in self._symbols_to_constraints.items()
            if len(symbols) > 0
        ]

    def _add_constraint(self, constraint: Boolean, inferred_by: Boolean):
        inferred_by_set = frozenset([inferred_by])
        symbols = _get_basic_symbols(constraint)
        self._symbols_to_constraints[symbols].add(
            _InferredConstraint(constraint, inferred_by_set)
        )
        for symbol in symbols:
            symbol_set = solveset(constraint, symbol, domain=_get_symbol_domain(symbol))
            if isinstance(symbol_set, Intersection):
                for subset in symbol_set.args:
                    if isinstance(subset, sympy.Set):
                        self._symbol_to_sets[symbol].add(
                            _InferredSet(subset, inferred_by)
                        )
                        self._add_possible_imputation_from_set(
                            subset, symbol, inferred_by_set
                        )
            elif isinstance(symbol_set, ConditionSet):
                # solveset couldn't return a simple set. Attempt to solve manually.
                # Is it a polynomial inequality?
                if (
                    isinstance(symbol_set.condition, Le | Ge | Gt | Lt)
                    and isinstance(symbol_set.condition.lts, Expr)
                    and isinstance(symbol_set.condition.gts, Expr)
                ):
                    expr = (
                        symbol_set.condition.lts - symbol_set.condition.gts
                        if (symbol in symbol_set.condition.lts.free_symbols)
                        ^ (symbol in symbol_set.condition.gts.free_symbols)
                        else None
                    )

                    if expr is not None and expr.is_polynomial(symbol):
                        # It is a polynomial inequality, solve it as such.
                        expr_poly = Poly(expr, symbol)
                        strict = isinstance(symbol_set.condition, Gt | Lt)
                        GreaterThan = (
                            (lambda x: Interval.Lopen(x, oo))
                            if strict
                            else (lambda x: Interval(x, oo))
                        )
                        LessThan = (
                            (lambda x: Interval.Ropen(-oo, x))
                            if strict
                            else (lambda x: Interval(-oo, x))
                        )
                        try:
                            expr_roots = roots(expr_poly, strict=True, multiple=True)
                            union_args = []
                            # The inequality is in the form of F > 0 if polynomial is negative, i.e., F is negative iff an even
                            # number of factors are negative. F < 0 if polynomial is postive, i.e. F is positive iff an odd
                            # number of factors are odd.
                            for number_of_negatives in range(
                                0 if expr_poly.LC() < 0 else 1,
                                len(expr_roots) + 1,
                                2,
                            ):
                                for negative_indexes in combinations(
                                    range(len(expr_roots)), number_of_negatives
                                ):
                                    intervals = {}
                                    for index, root in enumerate(expr_roots):
                                        for key in intervals.keys():
                                            quotient = key / root
                                            if (
                                                len(quotient.free_symbols) == 0
                                                and quotient.is_real
                                            ):
                                                intervals[key] = Intersection(
                                                    intervals[key],
                                                    LessThan(quotient)
                                                    if index in negative_indexes
                                                    else GreaterThan(quotient),
                                                )
                                                break
                                        else:
                                            intervals[root] = (
                                                LessThan(1)
                                                if index in negative_indexes
                                                else GreaterThan(1)
                                            )
                                    if all(
                                        isinstance(interval, Interval)
                                        for interval in intervals.values()
                                    ):
                                        union_args.append(
                                            Intersection(
                                                *(
                                                    Interval(
                                                        -oo
                                                        if interval.start == -oo
                                                        else root * interval.start,
                                                        oo
                                                        if interval.end == oo
                                                        else root * interval.end,
                                                        left_open=interval.left_open,
                                                        right_open=interval.right_open,
                                                    )
                                                    for root, interval in intervals.items()
                                                )
                                            )
                                        )
                            self._symbol_to_sets[symbol].add(
                                _InferredSet(Union(*union_args), inferred_by)
                            )
                        except UnsolvableFactorError:
                            pass
            else:
                self._symbol_to_sets[symbol].add(_InferredSet(symbol_set, inferred_by))
                self._add_possible_imputation_from_set(
                    symbol_set, symbol, inferred_by_set
                )

    def _add_possible_imputation_from_set(
        self, set_expr: sympy.Set, target_expr: Symbol, inferred_by: frozenset[Boolean]
    ):
        if isinstance(set_expr, FiniteSet) and len(set_expr) == 1:
            expr = set_expr.args[0]
            if isinstance(expr, Expr):
                self._imputations.append(
                    Imputation(
                        _get_basic_symbols(expr),
                        target_expr,
                        expr,
                        inferred_by=inferred_by,
                    )
                )

    @property
    def validations(self) -> list[Validation]:
        """Get all validation operations from the constraints.

        Returns
        -------
        list[Validation]
            Validation operations
        """
        return self._validations

    @property
    def imputations(self) -> list[Imputation]:
        """Get all imputation operations from the constraints.

        Returns
        -------
        list[Imputation]
            Imputation operations
        """
        return self._imputations



"""Operations module."""

from __future__ import annotations

from sympy import Symbol
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sympy.logic.boolalg import Boolean
    from sympy import Expr


@dataclass(eq=True, frozen=True)
class Operation:
    """Generic operation.

    Attributes
    ----------
    keys : frozenset[Symbol]
        Set of keys that this operation needs to check if its not missing before executing.
    inferred_by : frozenset[Boolean]
        Set of constraints which generated this operation.
    """

    keys: frozenset[Symbol]
    inferred_by: frozenset[Boolean] = field(kw_only=True, default_factory=frozenset)


@dataclass(eq=True, frozen=True)
class Validation(Operation):
    """Validation operation.

    Attributes
    ----------
    operations : frozenset[Boolean]
        A set of SymPy Boolean expressions that are used to validate the keys.
    """

    operations: frozenset[Boolean]

    def __str__(self):
        return (
            f'Validation: ({", ".join(str(k) for k in self.keys)}) => [{", ".join(str(op) for op in self.operations)}]'
            f' inferred by ({", ".join(str(op) for op in self.inferred_by)})'
        )

    def __repr__(self):
        return str(self)


@dataclass(eq=True, frozen=True)
class Imputation(Operation):
    """Imputation operation.

    Attributes
    ----------
    target_key : Symbol
        Key that will be imputed by this operation

    operation : Expr
        SymPy expression used to impute the target key
    """

    target_key: Symbol
    operation: Expr

    def __str__(self):
        return (
            f'Imputation: ({", ".join(str(k) for k in self.keys)}) => {str(self.target_key)} = {self.operation}'
            f' inferred by ({", ".join(str(op) for op in self.inferred_by)})'
        )

    def __repr__(self):
        return str(self)



# SPDX-FileCopyrightText: 2024-present Abdelrahman Abdelrahman <me@abogic.al>
#
# SPDX-License-Identifier: MIT
"""Validate and impute your dataset with mathematical expressions.

Symbolic Constraints, or `symconstraints` for short, allows you to express your dataset rules
using mathematical equations and expressions. It makes use of the powerful [SymPy](https://www.sympy.org) Computer Algebra System to analyze
mathematical expressions and infer all possible validation and imputation methods to your datasets. Helping you clean
your dataset with little code.

Examples
--------
Validate and impute a dataframe
>>> import pandas as pd
>>> from symconstraints import Constraints
>>> from symconstraints.pandas import symbols, check, set_invalid_all, impute
>>> from sympy import Eq
>>> df = pd.DataFrame(
...    {
...         "height": [5, 6, 8, 9],
...         "width": [3, 5, 7, None],
...         "area": [14, 30, None, 18],
...     },
...     dtype=float,
... )
>>> height, width, area = symbols(df, ["height", "width", "area"])
>>> constraints = Constraints([height > width, Eq(area, width * height)])
>>> check_result = check(constraints, df)
>>> df = set_invalid_all(check_result, df)
>>> df
    height  width  area
0     NaN    NaN   NaN
1     6.0    5.0  30.0
2     8.0    7.0   NaN
3     9.0    NaN  18.0
>>> imputed_df = impute(constraints, df)
>>> imputed_df
    height  width  area
0     NaN    NaN   NaN
1     6.0    5.0  30.0
2     8.0    7.0  56.0
3     9.0    2.0  18.0
"""

from symconstraints.constraints import Constraints, symbols
from symconstraints.operation import Operation, Validation, Imputation

__docformat__ = "numpy"
__all__ = ["Constraints", "symbols", "Operation", "Validation", "Imputation"]
